import application.Application;
import placement.OptimalRequestPlacement_Marginal;
import placement.OptimalRequestPlacement_Overall;
import placement.RequestPlacementOutput;
import system.Network;
import utils.Config;
import device.Device;

import java.io.IOException;
import java.util.*;

/**
 * Setup for Mixed Reality to test the placement of one request in different scenarios
 *
 * @author Klervie Toczé
 */
public class MixedReality_Optimization {

     public static void main(String[] args) {

        try {
            //Parameters for the different studies
            int numberOfFunctionInstances = 2;
            Boolean usingUniformLoadDistribution = false;
            int standardDeviation = 30;
            Boolean linksAreLoaded = false;
            Boolean oneCoreAllocation= true;
            Boolean randomStartLocation = true;
            Boolean colocationScenario = false; // Note: the colocation scenario is only run if the random start location is not activated

            String appId = "MR_Service"; // identifier of the application
            Application application = createApplication(appId);

            System.out.println("Description of the experiment:");
            System.out.println("- Number of functions instances: " + numberOfFunctionInstances);
            if (usingUniformLoadDistribution)
                {System.out.println("- Load distribution: UNIFORM");}
            else{
                System.out.println("- Load distribution: NORMAL with std " + standardDeviation);
            }
            System.out.println("   The (mean) load increases by 10% per scenario");
            if(linksAreLoaded){
                System.out.println("- Load on links: following a NORMAL distribution");
            }else{
                System.out.println("- Load on links: 0%");
            }
            if(oneCoreAllocation){
                System.out.println("- A function instance is allocated up to one CPU core");
            }else{
                System.out.println("- A function instance is allocated all remaining CPU capacity");
            }
            if (randomStartLocation)
            {System.out.println("- Start location: RANDOM ");}
            else{
                System.out.println("- Start location: FIXED");
            }

            //Determine startLocation = at which device the request is received. This is the placing device
            String startLocation;
            if (randomStartLocation){
                System.out.println("Random start location is decided per repetition");
                startLocation=null; //Will be set later in the random case
            }
            else{
                if(colocationScenario){
                    //For this scenario the start node is NOT random between scenarios
                    Integer randomStart=0; //Start device of 0 with the current function instance placement enables the possibility of colocation on the requesting device
                    startLocation= "gateway_" + randomStart;
                    System.out.println("Non random start location is "+startLocation);
                }
                else {
                    //For this scenario the start node is NOT random between scenarios
                    Integer randomStart=4;
                    startLocation= "gateway_" + randomStart;
                    System.out.println("Non random start location is "+startLocation);
                }

            }

            System.out.println("Creating the network to be used");
            Network network = new Network(application,startLocation,oneCoreAllocation);

            System.out.println("Placing the function instances");
            placeFunctionInstances(numberOfFunctionInstances,network.getEdgeDevices());

            System.out.println("Optimizing the placement of one request");

            //Below are the different optimization objectives
            OptimalRequestPlacement_Overall requestPlacementManager_obj1= new OptimalRequestPlacement_Overall(network);
            OptimalRequestPlacement_Marginal requestPlacementManager_obj2= new OptimalRequestPlacement_Marginal(network);

            //for each repetition
            for (int a = 0; a< Config.REPETITION_NUMBER; a++) {
                System.out.println("Repetition " + a);

                //For the random beginning device, we want to change it here
                if(randomStartLocation){
                    // Get new random beginning device
                    Random randomLoc = new Random(Config.SEED+a);
                    Integer randomStart=randomLoc.nextInt(11);
                    startLocation= "gateway_" + randomStart.toString();
                    System.out.println("Random start location is "+startLocation);
                    network.modifyPlacingDevice(startLocation);
                }

                //for each utilization scenario
                for (int i = 0; i < Config.LOAD_SCENARIO_NUMBER; i++) {
                    System.out.println("Utilization scenario " + i);

                    //Reset the timing/energy variables
                    network.clear();

                    //Create utilization data
                    network.createUtilizationData(i,a,usingUniformLoadDistribution,standardDeviation,linksAreLoaded);

                    //For every optimization objective to be tested
                    //Objective1 - Optimization Overall Energy
                    long startTiming = Calendar.getInstance().getTimeInMillis();
                    RequestPlacementOutput output_obj1 = requestPlacementManager_obj1.run(application, startLocation);
                    long endTiming = Calendar.getInstance().getTimeInMillis();
                    //Calculate and print how long it took to get the placement
                    long executionTime = endTiming - startTiming;
                    System.out.println("OVERALL OPT Placement request execution time: " + executionTime + " ms");

                    //Evaluate the placement according to the other metrics
                    network.evaluatePlacementOutput(output_obj1, startLocation, startLocation);

                    //Objective2 - Optimization Marginal Energy
                    startTiming = Calendar.getInstance().getTimeInMillis();
                    RequestPlacementOutput output_obj2 = requestPlacementManager_obj2.run(application, startLocation);
                    endTiming = Calendar.getInstance().getTimeInMillis();
                    //Calculate and print how long it took to get the placement
                    executionTime = endTiming - startTiming;
                    System.out.println("MARGINAL OPT Placement request execution time: " + executionTime + " ms");

                    //Evaluate the placement according to the other metrics
                    network.evaluatePlacementOutput(output_obj2, startLocation, startLocation);

                }
            }
        } catch (Exception e) {
            e.printStackTrace();
        }
    }

    private static Application createApplication(String appId) {

        Application application = Application.createApplication(appId);

        //Defining the different parts of the request function chain
        application.addAppLink("encodingModule_client", "decodingModule_edge", 200,250);
        application.addAppLink("decodingModule_edge", "pointcloudModule", 2000,500);
        application.addAppLink("pointcloudModule", "graphicoverlayModule", 1000,750);
        application.addAppLink("graphicoverlayModule", "encodingModule_edge", 200,500);
        application.addAppLink("encodingModule_edge", "decodingModule_client", 200, 250);

        //The request chain for the placement
        List<String> requestChain = new ArrayList<>();
        requestChain.add(0,"decodingModule_edge");
        requestChain.add(1,"pointcloudModule");
        requestChain.add(2,"graphicoverlayModule");
        requestChain.add(3,"encodingModule_edge");
        application.setRequestChain(requestChain);

        return application;
    }

    private static void placeFunctionInstances(int numberOfInstances,List<Device> edgeDevices){
        if(numberOfInstances>=2){
            for(Device device : edgeDevices){
                if(device.getName().equals("gateway_0")){
                    device.placeModuleOnDevice("decodingModule_edge");
                    device.placeModuleOnDevice("encodingModule_edge");
                }
                if(device.getName().equals("gateway_3")){
                    device.placeModuleOnDevice("decodingModule_edge");
                    device.placeModuleOnDevice("encodingModule_edge");
                }
                if(device.getName().equals("gateway_5")){
                    device.placeModuleOnDevice("pointcloudModule");
                }
                if(device.getName().equals("gateway_7")){
                    device.placeModuleOnDevice("pointcloudModule");
                }
                if(device.getName().equals("gateway_6")){
                    device.placeModuleOnDevice("graphicoverlayModule");
                }
                if(device.getName().equals("gateway_8")){
                    device.placeModuleOnDevice("graphicoverlayModule");
                }
            }
        }
        if(numberOfInstances>=4){
            for(Device device : edgeDevices){
                if(device.getName().equals("gateway_6")){
                    device.placeModuleOnDevice("decodingModule_edge");
                    device.placeModuleOnDevice("encodingModule_edge");
                }
                if(device.getName().equals("gateway_9")){
                    device.placeModuleOnDevice("decodingModule_edge");
                    device.placeModuleOnDevice("encodingModule_edge");
                }
                if(device.getName().equals("gateway_4")){
                    device.placeModuleOnDevice("pointcloudModule");
                }
                if(device.getName().equals("gateway_0")){
                    device.placeModuleOnDevice("pointcloudModule");
                }
                if(device.getName().equals("gateway_2")){
                    device.placeModuleOnDevice("graphicoverlayModule");
                }
                if(device.getName().equals("gateway_7")){
                    device.placeModuleOnDevice("graphicoverlayModule");
                }
            }
        }
        if(numberOfInstances>=6){
            for(Device device : edgeDevices){
                if(device.getName().equals("gateway_7")){
                    device.placeModuleOnDevice("encodingModule_edge");
                }
                if(device.getName().equals("gateway_5")){
                    device.placeModuleOnDevice("encodingModule_edge");
                }
                if(device.getName().equals("gateway_2")){
                    device.placeModuleOnDevice("decodingModule_edge");
                }
                if(device.getName().equals("gateway_4")){
                    device.placeModuleOnDevice("decodingModule_edge");
                }
                if(device.getName().equals("gateway_6")){
                    device.placeModuleOnDevice("pointcloudModule");
                }
                if(device.getName().equals("gateway_10")){
                    device.placeModuleOnDevice("pointcloudModule");
                }
                if(device.getName().equals("gateway_0")){
                    device.placeModuleOnDevice("graphicoverlayModule");
                }
                if(device.getName().equals("gateway_9")){
                    device.placeModuleOnDevice("graphicoverlayModule");
                }
            }
        }
    }

}