package placement;

import application.Application;
import device.Device;
import net.sf.javailp.*;
import solver.SolverGurobi;
import org.apache.commons.math3.util.Pair;
import system.Network;
import utils.Triple;
import utils.Utils;

import java.util.*;

/**
 * Created by Klervie Toczé
 * Request Placement using ILP
 * Optimization objective: Marginal Energy
 */
public class OptimalRequestPlacement_Marginal implements RequestPlacementLogic {

    Network network;
    List<Device> edgeDevicesInArea;//edge devices that can be considered for placement
    Map<String,Device> edgeDevicesInAreaByName;

    private String startLocation; //The device name of the start location for the request
    private String destinationLocation; //The device name of the destination location for the request

    public OptimalRequestPlacement_Marginal(Network network) {
        this.network=network;
        this.edgeDevicesInArea=network.getEdgeDevices();
        this.edgeDevicesInAreaByName = new HashMap<>();
        for (Device device : edgeDevicesInArea){
            this.edgeDevicesInAreaByName.put(device.getName(),device);
        }
    }

    private Triple<List<String>,List<String>,List<String>> optimizationSolve(List<String> requestChain){

        SolverFactory factory = new SolverFactoryGurobi();
        factory.setParameter(Solver.VERBOSE,0);
        factory.setParameter(Solver.TIMEOUT,1000); //Timeout of 1000 seconds

        //Get the names of the function instances
        List<String> functionInstancesLongNames = new ArrayList<>();
        List<String> functionInstancesShortNames = new ArrayList<>();
        functionInstancesShortNames.add(Utils.START);
        functionInstancesLongNames.add(Utils.START);
        for (Device device : edgeDevicesInArea){
            if(device.getDeviceType()== Device.EDGE) {
                //We are not interested by adding end device nodes
                //The function instances placed are accessible via Device
                    for (String module:device.getPlacedModulesOnDevice()){
                        //Note: the dash in between is to facilitate re-split later
                        String longName= device.getName() +"-" + module;
                        //Create a shorter name - Assuming a node name of type "gateway_0", we want the first character and the number
                        // followed by the two first letters of the module name
                        String shortNameDevice;
                        String nodeName=device.getName();
                        if (nodeName.charAt(nodeName.length()-2)!="_".charAt(0)){
                            shortNameDevice= ""+nodeName.charAt(0)+nodeName.charAt(nodeName.length()-2)+nodeName.charAt(nodeName.length()-1);
                        }
                        else {
                            shortNameDevice= ""+nodeName.charAt(0)+nodeName.charAt(nodeName.length()-1);
                        }
                        String shortName = shortNameDevice +module.charAt(0)+module.charAt(1);
                        functionInstancesShortNames.add(shortName);
                        functionInstancesLongNames.add(longName);
                    }
            }
        }
        functionInstancesShortNames.add(Utils.END);
        functionInstancesLongNames.add(Utils.END);

        //Create a shorter version of the names for the requestChain
        List<String> shortNameRequestChain = new ArrayList<>();
        for (String longName : requestChain){
            String shortName = ""+longName.charAt(0)+longName.charAt(1)+longName.charAt(2);
            shortNameRequestChain.add(shortName);
        }

        //Input variables
        //Create/Get the coefficients for the lambdas values in the objective function
        int beta = requestChain.size()+2;//Add to be +2 to be number of links between instance + 1 due to constraint 6
        double deadline = 100;

        Problem problem = new Problem();

        //Decision variables
        for (int i=0;i<functionInstancesShortNames.size();i++){
            for (int j=0;j<functionInstancesShortNames.size();j++){
                    problem.setVarType("x" + functionInstancesShortNames.get(i) + functionInstancesShortNames.get(j), Boolean.class);
            }
        }
        for (int i=0;i<functionInstancesShortNames.size();i++){
                problem.setVarType("y"+functionInstancesShortNames.get(i), Boolean.class);
        }
        for (int i=0;i<functionInstancesShortNames.size();i++){
            for (int j=0;j<functionInstancesShortNames.size();j++){
                    problem.setVarType("o" + functionInstancesShortNames.get(i) + functionInstancesShortNames.get(j), Integer.class);
                    problem.setVarLowerBound("o"+functionInstancesShortNames.get(i)+functionInstancesShortNames.get(j),0);
                    problem.setVarUpperBound("o"+functionInstancesShortNames.get(i)+functionInstancesShortNames.get(j),beta);
            }
        }

        //Create the objective constraint
        Linear linear = new Linear();
        //Travelling part
        for (int i=0;i<functionInstancesShortNames.size();i++){
            for (int j=0;j<functionInstancesShortNames.size();j++){
                    linear.add(network.getLinkMarginalEnergyConsumption(functionInstancesLongNames.get(i),functionInstancesLongNames.get(j),this.startLocation,this.destinationLocation), "x"+functionInstancesShortNames.get(i)+functionInstancesShortNames.get(j));
            }
        }
        //Purchasing part
        for (int i=1;i<functionInstancesShortNames.size()-1;i++){ //Exclude the start and end nodes
                linear.add(network.getInstanceMarginalEnergyConsumption(functionInstancesLongNames.get(i)), "y"+functionInstancesShortNames.get(i));
        }
        problem.setObjective(linear, OptType.MIN);

        //Add the constraints
        //Number 6
        //Special case source
        linear = new Linear();
        for (int j=0;j<functionInstancesShortNames.size();j++){//w
            linear.add(1, "x"+functionInstancesShortNames.get(j)+functionInstancesShortNames.get(0));
        }
        for (int j=0;j<functionInstancesShortNames.size();j++){//z
            linear.add(-1, "x"+functionInstancesShortNames.get(0)+functionInstancesShortNames.get(j));
        }
        problem.add("6-source",linear, "=", -1);
        //Special case destination
        linear = new Linear();
        for (int j=0;j<functionInstancesShortNames.size();j++){//w
            linear.add(1, "x"+functionInstancesShortNames.get(j)+functionInstancesShortNames.get(functionInstancesShortNames.size()-1));
        }
        for (int j=0;j<functionInstancesShortNames.size();j++){//z
            linear.add(-1, "x"+functionInstancesShortNames.get(functionInstancesShortNames.size()-1)+functionInstancesShortNames.get(j));
        }
        problem.add("7-dest",linear, "=", 1);
        //Regular case
        linear = new Linear();
        for (int i=1;i<functionInstancesShortNames.size()-1;i++){ //v minus source and destination
            for (int j=0;j<functionInstancesShortNames.size();j++){//w
                linear.add(1, "x"+functionInstancesShortNames.get(j)+functionInstancesShortNames.get(i));
            }
            for (int j=0;j<functionInstancesShortNames.size();j++){//z
                linear.add(-1, "x"+functionInstancesShortNames.get(i)+functionInstancesShortNames.get(j));
            }
            problem.add("5-regular"+i,linear, "=", 0);
            linear = new Linear(); //Do not forget to reset!
        }
        //Number 8
        linear = new Linear();
        for (int i=0;i<requestChain.size();i++){ //f
            for (int j=1;j<functionInstancesShortNames.size()-1;j++){//v [(should not place on the source or the end)]
                linear.add(isFunctionInstanceForFunction(functionInstancesLongNames.get(j),requestChain.get(i)), "y"+functionInstancesShortNames.get(j));
            }
            problem.add("8_"+i,linear, "=", 1);
            linear = new Linear();
        }

        //Number 9
        linear = new Linear();
            for (int j=1;j<functionInstancesShortNames.size()-1;j++){//v - source - destination
                for (int k=0;k<functionInstancesShortNames.size();k++) {//w
                    linear.add(1, "x" + functionInstancesShortNames.get(j) + functionInstancesShortNames.get(k));
                }
                linear.add(-1, "y" + functionInstancesShortNames.get(j));
                problem.add("9_"+j,linear, "=",0);
                linear = new Linear();
            }
        //Number 10
        for (int i=0;i<functionInstancesShortNames.size();i++){ //w
            linear = new Linear();
            for (int j=0;j<functionInstancesShortNames.size();j++){//v
                linear.add(1, "x"+functionInstancesShortNames.get(j)+functionInstancesShortNames.get(i));
            }
            problem.add("10_"+i,linear, "<=", 1);
        }
        //Number 11
        for (int i=0;i<functionInstancesShortNames.size();i++){ //v
            linear = new Linear();
            for (int j=0;j<functionInstancesShortNames.size();j++){//w
                linear.add(1, "x"+functionInstancesShortNames.get(i)+functionInstancesShortNames.get(j));
            }
            problem.add("11_"+i,linear, "<=", 1);
        }

        //Number 12
        linear = new Linear();
        for (int i=0;i<functionInstancesShortNames.size();i++){ //v
            for (int j=0;j<functionInstancesShortNames.size();j++){//w
                linear.add(1, "o"+functionInstancesShortNames.get(i)+functionInstancesShortNames.get(j));
                linear.add(-beta, "x"+functionInstancesShortNames.get(i)+functionInstancesShortNames.get(j));
                problem.add("12_"+i+"_"+j,linear, "<=", 0);
                linear = new Linear();
            }
        }
        //Number 13 - special case source node
        linear = new Linear();
        for (int j=0;j<functionInstancesShortNames.size();j++) {//w
            linear.add(1, "o"+functionInstancesShortNames.get(j)+functionInstancesShortNames.get(0));
            linear.add(-1, "o"+functionInstancesShortNames.get(0)+functionInstancesShortNames.get(j));
            linear.add(-1, "x"+functionInstancesShortNames.get(0)+functionInstancesShortNames.get(j));
        }
        problem.add("13-source",linear, "=", -beta);
        // Number 13
        linear = new Linear();
        for (int i=1;i<functionInstancesShortNames.size()-1;i++) { //v - source - destination
            for (int j=0;j<functionInstancesShortNames.size();j++) {//w
                linear.add(1, "o"+functionInstancesShortNames.get(j)+functionInstancesShortNames.get(i));
                linear.add(-1, "o"+functionInstancesShortNames.get(i)+functionInstancesShortNames.get(j));
                linear.add(-1, "x"+functionInstancesShortNames.get(i)+functionInstancesShortNames.get(j));
            }
            problem.add("13_"+i,linear, "=", 0);
            linear = new Linear();
        }
        //Number 14 - My alternative of constraint to enforce order
        linear = new Linear();
        for (int i=1;i<functionInstancesShortNames.size()-1;i++){ //v (- destination) - source (cannot place on source anyway)
            for (int j=1;j<functionInstancesShortNames.size()-1;j++){ //w (- destination) - source
                if (i != j) {
                    for (int k = 0; k < requestChain.size(); k++) {
                        for (int l = 0; l < requestChain.size(); l++) {
                            if(k!=l){
                                //The order in the request chain order should be the order in which the function should be placed
                                if(k==l-1){
                                    if(isFunctionInstanceForFunction(functionInstancesLongNames.get(i),requestChain.get(k))==1 &&isFunctionInstanceForFunction(functionInstancesLongNames.get(j),requestChain.get(l)) ==1) {
                                        for (int m = 0; m < functionInstancesShortNames.size()-1; m++) {//alpha
                                            linear.add(isFunctionInstanceForFunction(functionInstancesLongNames.get(i),requestChain.get(k)), "o" + functionInstancesShortNames.get(m) + functionInstancesShortNames.get(i));
                                            linear.add(-isFunctionInstanceForFunction(functionInstancesLongNames.get(j),requestChain.get(l)) , "o" + functionInstancesShortNames.get(m) + functionInstancesShortNames.get(j));
                                            linear.add(-beta, "x" + functionInstancesShortNames.get(m)+functionInstancesShortNames.get(j));
                                        }
                                        linear.add(-1, "x" + functionInstancesShortNames.get(i)+functionInstancesShortNames.get(j));
                                        linear.add(-beta, "y" + functionInstancesShortNames.get(i));
                                        linear.add(beta, "y" + functionInstancesShortNames.get(j));
                                        problem.add("14new_" + i + "_" + j + "_" + k + "_" + l, linear, ">=", -beta);
                                        linear = new Linear();
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }

        //Number 15
        linear = new Linear();
        //Travelling part
        for (int i=0;i<functionInstancesShortNames.size();i++){
            for (int j=0;j<functionInstancesShortNames.size();j++){
                    linear.add(network.getLinkTransmissionTime(functionInstancesLongNames.get(i),functionInstancesLongNames.get(j)), "x"+functionInstancesShortNames.get(i)+functionInstancesShortNames.get(j));
            }
        }
        //Purchasing part
        for (int i=1;i<functionInstancesShortNames.size()-1;i++){ //Exclude the start node and the end node
            String currentName = functionInstancesLongNames.get(i);
            Pair<String, String> nameSplit= Utils.splitLongFunctionInstanceName(currentName);
                linear.add(network.getExecutionTime(nameSplit.getKey(),nameSplit.getValue()), "y"+functionInstancesShortNames.get(i));
        }
        problem.add("15",linear, "<=", deadline);

        //No self loop
        linear = new Linear();
        for (int i=0;i<functionInstancesShortNames.size();i++){
                linear.add(1, "x"+functionInstancesShortNames.get(i)+functionInstancesShortNames.get(i));
        }
        problem.add("noSelfLoop",linear, "=", 0);

        //Force so that nothing is placed on start and end
        linear = new Linear();
        linear.add(1, "y"+functionInstancesShortNames.get(0));
        problem.add("NoPlacementOnStart",linear, "=", 0);
        linear = new Linear();
        linear.add(1, "y"+functionInstancesShortNames.get(functionInstancesShortNames.size()-1));
        problem.add("NoPlacementOnEnd",linear, "=", 0);

        //Solve the problem
        Solver solver = new SolverGurobi();
        Result result = solver.solve(problem);

        if (result == null){
            //Infeasible model, the request most likely cannot be served in time and should be dropped to avoid wasting resources
            System.out.println("The request is infeasible");
            return null;
        }

        System.out.println(result);

        //Transform the solver output into a placement variable
        List<String> optimalTourOrder = new LinkedList<>();
        List<String> optimalTourVisitReason = new LinkedList<>();
        List<String> allElementsForTiming = new LinkedList<>();

        //Get which links were selected by the solver
        Map<String,String> selectedLinks =new HashMap<>();
        for (int i=0;i<functionInstancesShortNames.size();i++){
            for (int j=0;j<functionInstancesShortNames.size();j++){
                //Check all the decision variables x values and extract the ones selected
                if (result.getPrimalValue("x"+functionInstancesShortNames.get(i)+functionInstancesShortNames.get(j)).intValue()==1){
                   selectedLinks.put(functionInstancesLongNames.get(i),functionInstancesLongNames.get(j));
                   allElementsForTiming.add(functionInstancesLongNames.get(i)+"=>"+functionInstancesLongNames.get(j)); //Add all links
                }
            }
        }
        //Follow the path and fill the pre-return variables
        //Deal with the start device
        optimalTourOrder.add(startLocation);
        optimalTourVisitReason.add("Start");
        String currentInstance = selectedLinks.get(Utils.START);
        //Deal with the middle devices based on the solver output
        while(!currentInstance.equals(Utils.END)){
            allElementsForTiming.add(currentInstance);
            Pair<String,String> instanceSplit = Utils.splitLongFunctionInstanceName(currentInstance);
            optimalTourOrder.add(instanceSplit.getKey());
            optimalTourVisitReason.add(instanceSplit.getValue());
            currentInstance=selectedLinks.get(currentInstance);
        }

        //Deal with the destination device
        optimalTourOrder.add(destinationLocation);
        optimalTourVisitReason.add("Destination");

        //Creating the return variable
        Triple<List<String>,List<String>,List<String>> placement= new Triple<>(optimalTourOrder,optimalTourVisitReason, allElementsForTiming);
        return placement;
    }

    public RequestPlacementOutput run(Application currentApplication, String startLocation) {
        //Getting the start location, i.e. the device at which the tuple is received
        this.startLocation = startLocation;
        this.destinationLocation = startLocation;//In the current implementation, start and end device are the same.

        RequestPlacementOutput placement = new RequestPlacementOutput();
        List<String> requestChain = currentApplication.getRequestChain();

        Triple<List<String>,List<String>,List<String>> output = optimizationSolve(requestChain);
        if (output == null){
            //No placement solution
            return null;
        }

        //Fix the placement class
        placement.setObtainedTour(output.getLeft());
        placement.setObtainedTourVisitReason(output.getMiddle());
        placement.updateAppFunctionToDevice();
        placement.setElementsForTiming(output.getRight());

        return placement;
    }

    private int isFunctionInstanceForFunction(String longFunctionInstanceName,String functionName){
        //Note: we assume that the long name is "device name" +"-"+"functionName"
        //If it is not like that, this function does not work
        boolean answer =longFunctionInstanceName.contains(functionName);
        return (answer) ? 1 : 0;
    }

}
